#include "Connect4.h"

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "random.h"

Connect4 *connect4_create(int32_t mode, int32_t rows, int32_t cols) {
    Connect4 *connect4 = (Connect4 *)malloc(sizeof(Connect4));
    connect4->mode = mode;
    connect4->current_player = PiecePlayer1;
    connect4->rows = rows;
    connect4->cols = cols;
    connect4->grid = (Piece **)malloc(sizeof(Piece *) * rows);

    for (int32_t y = 0; y < connect4->rows; y += 1) {
        connect4->grid[y] = (Piece *)malloc(sizeof(Piece) * cols);

        for (int32_t x = 0; x < connect4->cols; x += 1) {
            connect4->grid[y][x] = PieceNone;
        }
    }

    return connect4;
}

void connect4_destroy(Connect4 *connect4) {
    for (int32_t y = 0; y < connect4->rows; y += 1) {
        free(connect4->grid[y]);
    }

    free(connect4->grid);
    free(connect4);
}

Connect4 *connect4_copy(Connect4 *connect4) {
    Connect4 *copied_connect4 = connect4_create(connect4->mode, connect4->rows, connect4->cols);

    for (int32_t y = 0; y < connect4->rows; y += 1) {
        for (int32_t x = 0; x < connect4->cols; x += 1) {
            copied_connect4->grid[y][x] = connect4->grid[y][x];
        }
    }

    return copied_connect4;
}

void print_grid_separator(int32_t n, char *front, char *rear, char *separator1, char *separator2) {
    printf("%s", front);

    for (int32_t i = 0; i < n; i += 1) {
        printf("%s", separator1);

        if (i < n - 1) {
            printf("%s", separator2);
        }
    }

    printf("%s", rear);
    printf("\n");
}

void print_grid_body(Piece **grid, int32_t rows, int32_t cols, int32_t y) {
    char cell_dict[3];
    cell_dict[PieceNone] = ' ';
    cell_dict[PiecePlayer1] = 'X';
    cell_dict[PiecePlayer2] = 'O';

    for (int32_t x = 0; x < cols; x += 1) {
        printf("│%c", cell_dict[grid[y][x]]);
    }

    printf("│");
    printf("\n");

    if (y < rows - 1) {
        print_grid_separator(cols, "├", "┤", "─", "┼");
    }
}

void print_grid(Piece **grid, int32_t rows, int32_t cols) {
    print_grid_separator(cols, "┌", "┐", "─", "┬");

    for (int32_t y = 0; y < rows; y += 1) {
        print_grid_body(grid, rows, cols, y);
    }

    print_grid_separator(cols, "└", "┘", "─", "┴");
}

void connect4_print(Connect4 *connect4) {
    print_grid(connect4->grid, connect4->rows, connect4->cols);

    for (int32_t x = 0; x < connect4->cols; x += 1) {
        printf(" %d", x + 1);
    }

    printf("\n");
}

bool connect4_is_full(Connect4 *connect4) {
    for (int32_t y = 0; y < connect4->rows; y += 1) {
        for (int32_t x = 0; x < connect4->cols; x += 1) {
            if (connect4->grid[y][x] == PieceNone) {
                return false;
            }
        }
    }

    return true;
}

int32_t count_contiguous_pieces(Connect4 *connect4, int32_t x, int32_t y, int32_t x_direction, int32_t y_direction, Piece piece) {
    if (x < 0 || x >= connect4->cols || y < 0 || y >= connect4->rows || connect4->grid[y][x] != piece) {
        return 0;
    }

    return 1 + count_contiguous_pieces(connect4, x + x_direction, y + y_direction, x_direction, y_direction, piece);
}

bool connect4_check_win(Connect4 *connect4, int32_t x, int32_t y, Piece piece) {
    bool win = false;

    win |= (count_contiguous_pieces(connect4, x, y, -1, 0, piece) + count_contiguous_pieces(connect4, x, y, 1, 0, piece) - 1) == 4;
    win |= count_contiguous_pieces(connect4, x, y, 0, 1, piece) == 4;
    win |= (count_contiguous_pieces(connect4, x, y, 1, -1, piece) + count_contiguous_pieces(connect4, x, y, -1, 1, piece) - 1) == 4;
    win |= (count_contiguous_pieces(connect4, x, y, -1, -1, piece) + count_contiguous_pieces(connect4, x, y, 1, 1, piece) - 1) == 4;

    return win;
}

PlacingStatus connect4_place_piece(Connect4 *connect4, Piece piece, int32_t x) {
    for (int32_t y = connect4->rows - 1; y >= 0; y -= 1) {
        if (connect4->grid[y][x] == PieceNone) {
            connect4->grid[y][x] = piece;

            if (connect4_check_win(connect4, x, y, piece)) {
                return PlacingStatusPlayerWon;
            }

            if (connect4_is_full(connect4)) {
                return PlacingStatusGridFull;
            }

            return PlacingStatusPiecePlaced;
        }
    }

    return PlacingStatusColumnFull;
}

Piece change_player(Piece current_player) {
    if (current_player == PiecePlayer1) {
        return PiecePlayer2;
    }

    return PiecePlayer1;
}

PlacingStatus connect4_play_random(Connect4 *connect4, void (*on_piece_placed)(Connect4 *)) {
    PlacingStatus status;

    do {
        status = connect4_play(connect4, random_range(1, connect4->cols), on_piece_placed);
    } while (status == PlacingStatusColumnFull && status == PlacingStatusInvalidColumn);

    return status;
}

PlacingStatus connect4_play_smart(Connect4 *connect4, void (*on_piece_placed)(Connect4 *)) {
    Piece current_player = connect4->current_player;

    for (int32_t i = 0; i < 2; i += 1) {
        for (int32_t x = 0; x < connect4->cols; x += 1) {
            Connect4 *copied_connect4 = connect4_copy(connect4);
            PlacingStatus status = connect4_place_piece(copied_connect4, current_player, x);

            if (status == PlacingStatusColumnFull) {
                connect4_destroy(copied_connect4);
                continue;
            }

            int32_t y = 0;

            for (int32_t i = 0; i < copied_connect4->rows; i += 1) {
                if (copied_connect4->grid[i][x] != PieceNone) {
                    y = i;
                    break;
                }
            }

            if (connect4_check_win(copied_connect4, x, y, current_player)) {
                connect4_destroy(copied_connect4);
                return connect4_play(connect4, x + 1, on_piece_placed);
            }

            connect4_destroy(copied_connect4);
        }

        current_player = change_player(current_player);
    }

    return connect4_play_random(connect4, on_piece_placed);
}

PlacingStatus connect4_play(Connect4 *connect4, int32_t x, void (*on_piece_placed)(Connect4 *)) {
    if (x < 1 || x > connect4->cols) {
        return PlacingStatusInvalidColumn;
    }

    PlacingStatus status = connect4_place_piece(connect4, connect4->current_player, x - 1);

    if (status != PlacingStatusInvalidColumn || status != PlacingStatusColumnFull) {
        on_piece_placed(connect4);
    }

    if (status != PlacingStatusPiecePlaced) {
        return status;
    }

    connect4->current_player = change_player(connect4->current_player);

    if (connect4->mode != 3 && connect4->current_player == PiecePlayer2) {
        if (connect4->mode == 1) {
            status = connect4_play_random(connect4, on_piece_placed);
        } else if (connect4->mode == 2) {
            status = connect4_play_smart(connect4, on_piece_placed);
        }
    }

    return status;
}
